/**********************************************************************/
/*  PROGRAM MPISERIAL

PURPOSE: This program is a simple MPI wrapper that runs a serial
program on each MPI rank.  It has been tested on GAEA, Cirrus,
Stratus, Tide, Gyre, Zeus and Jet.  Each MPI rank has two variables
set: SCR_COMM_RANK and SCR_COMM_SIZE containing the MPI rank and
communicator size.  

CAVEATS: This program is not a substitute for a vendor MPMD interface
that provides support for MPI in the commands, in that it can only
handle purely serial programs.  The script that is executed cannot run
any MPI programs, or most MPI implementations (including those on
Zeus, GAEA, Tide and Gyre) will fail in generally unpredictable ways.
The coupled model case is an example of what must be handled by the
vendor's MPMD method.

AUTHORSHIP: 

 circa 2010-2012 - various implementations made by Sam
   Trahan, George Vandenberghe and Hendrik Tollman.  

 late 2012 - Sam Trahan made a version that included best features of
   all versions

 early 2013 - wrong version (old wave model version) given to NCO,
   lacking error handling and SPMD support

 2013 March - "Merged" version tested on Tide, Gyre, GAEA, Zeus, Jet,
   Cirrus and Stratus and finalized.

 2015 April - Sam Trahan added optional verbosity and a workaround for
   an LSF signal handling issue

 2016 April - Sam trahan replaced system() with vfork and execve to
   allow immediate exit on error.

---------

CALLING CONVENTION: This program can be called in one of two different
ways:

SINGLE PROGRAM MULTIPLE DATA (SPMD)

  mpiexec mpiserial 'command to run'

Called like that, the program will run the specified command on all
MPI ranks using /bin/sh.  This is intended to be used to execute
simple commands like hostname or sync.  However, more complex programs
can differentiate between tasks using $SCR_COMM_RANK.

Multiple arguments can be specified in SPMD mode, but they will simply
be appended to one another with a space between each.  That behavior
may be changed eventually to correctly handle arguments via the execvp
family of functions.


MULTIPLE PROGRAM MULTIPLE DATA (MPMD)

  cp /my/command/file cmdfile
  mpiexec mpiserial

OR

  export SCR_CMDFILE=/my/command/file
  mpiexec mpiserial

The program will read the first N lines of the command file, where N
is the size of MPI_COMM_WORLD.  Each rank M runs the command on line
M+1 (keeping in mind that MPI ranks are zero-based).  If the command
file is too short, extra ranks will run "/bin/true".

---------

EXIT STATUS: The return status from mpiserial is 0 if the serial
programs all exited with status 0.  If a command could not be
executed, or if an MPI error is encountered, mpiserial will return a
non-zero status.

 */
/**********************************************************************/

#include <time.h>
#include <sys/time.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>

extern char **environ; /* special unix global for environment vars */

static int default_immediate_exit = 0; /* flag: immediate exit on error? */
static int default_verbosity = 1; /* default verbosity level */

static const int max_command_len=1048576; /* Limit command length to avoid buffer overflow attacks */

static int mpi_inited=0, mpi_rank=-99, verbosity=0;
static const char * const  default_cmdfile="cmdfile"; /* default cmdfile name */
static const char * const  special_non_command="***"; /* don't run a command, exit 0 */

void message(int level,const char *format,...) {
  /* Debug message function.  Will print the message to stderr only if
     level<=verbosity. */
  va_list ap;

  va_start(ap,format);
  if(level<=verbosity)
    vfprintf(stderr,format,ap);
  va_end(ap);
}

const char *timestamp() {
  static char timebuf[100];
  struct timeval nowmilli;
  struct tm now;
  size_t len;

  memset(timebuf,0,100);
  
  gettimeofday(&nowmilli,NULL);
  gmtime_r(&(nowmilli.tv_sec),&now);
  
  strftime(timebuf,100,"%m/%d %H:%M:%S",&now);
  len=strlen(timebuf);
  snprintf(timebuf+len,100-len,".%03d",nowmilli.tv_usec/1000);
  
  return timebuf;
}

void die(const char *format,...) {
  /* Error handling function.  Exits program with a message, calling
     MPI_Abort if needed.  On some MPI implementations it is critical
     to call MPI_Abort or the program may hang. */
  va_list ap;

  va_start(ap,format);
  vfprintf(stderr,format,ap);
  va_end(ap);

  if(mpi_inited)
    MPI_Abort(MPI_COMM_WORLD,2);
  exit(2);
  abort();
}

void mpicall(int loglev,int ret,const char *name) {
  /* This is the MPI error handling function.  It is a simple wrapper
     around "die" that generates an intelligible error message */
  char error[MPI_MAX_ERROR_STRING+1]="";
  int len=-1,err2;
  if(ret!=MPI_SUCCESS) {
    if((err2=MPI_Error_string(ret,error,&len))==MPI_SUCCESS) {
      error[len]='\0';
      die("%s: MPI error %d when %s: %s\n",timestamp(),ret,name,error);
    } else {
      die("%s: MPI error %d when %s, and MPI_Error_string could not process "
	  "that error number (it said %d) so I don't even know what went wrong!\n",
          timestamp(),ret,name,err2);
    }
  }
  if(mpi_rank<0)
    message(loglev,"%s: MPI Serial: %s successful.\n",timestamp(),name);
  else
    message(loglev,"%s: MPI Serial %d: %s successful.\n",timestamp(),mpi_rank,name);
}

int scr_immediate_exit(void) {
  char *immediate_exits=getenv("SCR_IMMEDIATE_EXIT");;
  int immediate_exiti=default_immediate_exit;
  if(!immediate_exits || !*immediate_exits)
    return default_immediate_exit;
  if(!strcasecmp(immediate_exits,"yes") || !strcasecmp(immediate_exits,"true"))
    return 1;
  if(!strcasecmp(immediate_exits,"no") || !strcasecmp(immediate_exits,"false"))
    return 0;
  if(sscanf(immediate_exits,"%d",&immediate_exiti)>=1)
    return immediate_exiti;
  return default_immediate_exit;
}

int scr_verbosity(void) {
  char *cverb=getenv("SCR_VERBOSITY");;
  int verbosity=default_verbosity;
  if(!cverb) 
    return verbosity;
  sscanf(cverb,"%d",&verbosity); /* on failure, sscanf will not modify verbosity */
  return verbosity;
}

int is_spmd(const int argc,const int rank) {
  /* Decides if the program is called in SPMD mode.  Returns 1 for
     SPMD, 0 for MPMD and aborts if unsure. */
  int spmd,mpmd;
  char *pgmmodel;

  if(rank==0) {
    spmd=(argc>1);
    mpmd=(!spmd);
    pgmmodel=getenv("SCR_PGMMODEL");
    if(pgmmodel) {
      if(!strcasecmp(pgmmodel,"MPMD")) {
        spmd=0 ; mpmd=1;
      } else if(!strcasecmp(pgmmodel,"SPMD")) {
        spmd=1 ; mpmd=0;
      } else {
        die("%s: Invalid value \"%s\" for $SCR_PGMMMODEL\n",
	    timestamp(),pgmmodel);
      }
    }
  }
  mpicall(2,MPI_Bcast(&spmd,1,MPI_INTEGER,0,MPI_COMM_WORLD),"calling MPI_Bcast");
  return spmd;
}

int haveenv(const char *c) {
  /* Returns 1 if the specified variable exists and consists of more
     than just blank characters.  Returnsn 0 otherwise. */
  char *val=getenv(c);
  if(!val || !*val)
    return 0;
  while(*val) {
    if(!isblank(*val))
      return 1;
    val++;
  }
  return 0;
}

void lsf_signal_workaround() {
  int isig;
  int lsf=0;
  
  /* Decide if we are using LSF */
  lsf+=haveenv("LSB_JOBNAME");
  lsf+=haveenv("LSB_HOSTS");
  lsf+=haveenv("LSB_JOBINDEX");
  lsf+=haveenv("LSB_SUB_USER");
  lsf+=haveenv("LSB_TRAPSIGS");
  lsf+=haveenv("LSF_VERSION");
  lsf+=haveenv("LSF_INVOKE_CMD");
  lsf+=haveenv("LSF_BINDIR");
  lsf=(lsf>=2); /* Require at least two recognized variables */
  
  if(!lsf) {
    if(mpi_rank>=0)
      message(2,"%s: MPI Serial %d: not using LSF, retain signal handling setup\n",timestamp(),mpi_rank);
    else
      message(2,"%s: MPI Serial: not using LSF, retain signal handling setup\n",timestamp());
  }

  if(mpi_rank>=0)
    message(2,"%s: MPI Serial %d: restoring signal handlers 1-63 to work around LSF bug\n",timestamp(),mpi_rank);
  else
    message(2,"%s: MPI Serial: restoring signal handlers 1-63 to work around LSF bug\n",timestamp());

  for(isig=1;isig<=63;isig++)
    signal(isig,SIG_DFL);
}

char *append_args(const int argc,const char **argv) {
  /* Appends contents of argv together with spaces in between,
     allocating memory as needed.  Returns the resulting string. */
  char *buf;
  int iarg;
  size_t len,len1,buflen;

  len=0;
  for(iarg=1;iarg<argc;iarg++)
    len+=1+strlen(argv[iarg]);
  if(!(buf=(char*)malloc(len)))
    die("%s: Unable to allocate %llu bytes.\n",
	timestamp(),(unsigned long long)len);

  buflen=len;
  len=0;
  for(iarg=1;iarg<argc;iarg++) {
    len1=strlen(argv[iarg]);
    memcpy(buf+len,argv[iarg],len1);
    len+=len1;

    if(iarg==argc-1)
      buf[len]='\0';
    else
      buf[len]=' ';
    len++;
  }
  if(len!=buflen)
    die("%s: ASSERTION FAILURE: %llu!=%llu in %s:%d\n",
	timestamp(),
        (unsigned long long)len,
        (unsigned long long)buflen,
        __FILE__,__LINE__);

  return buf;
}

double doublesleep(double s) {
  /* Sleep s nanoseconds, return unslept time if sleep was
     interrupted. */
  struct timespec rem, req={ floor(s), 1e9*(s-floor(s)) };
  nanosleep(&req,&rem);
  return rem.tv_sec + rem.tv_nsec*1e-9;
}

void checkpoint_message(int verbosity,int rc,int running,int max_rc,int max_running,pid_t pid) {
  if(running)
    message(verbosity,"%s: MPI Serial %d: process %lld is still running.\n",
	    timestamp(),mpi_rank,(long long)pid);
  else if(max_running)
    message(verbosity,"%s: MPI Serial %d: process %lld exited (status %d), waiting"
	    " for other processes.\n",timestamp(),mpi_rank,(long long)pid,rc);
  else
    message(verbosity,"%s: MPI Serial %d: all processes exited.  My process (%lld) "
	    "status was %d, max status was %d\n",
	    timestamp(),mpi_rank,(long long)pid,rc,max_rc);
}


unsigned run(const char *command) {
  /* Runs the specified command, and returns an exit status which will
     be in the range 0-255.  This correctly handles killed or stopped
     jobs, returning 128 plus the signal number. */
  unsigned rc=99, running=1, checked;
  unsigned send[3],recv[3];
  int ret, immediate_exit, maxret, i;
  time_t lastprint=0,now=0;
  pid_t pid,pid2;
  char *cmd[4];

  immediate_exit=scr_immediate_exit();

  if(!strcmp(command,special_non_command))
    return 0;

  cmd[0]=strdup("/bin/sh");
  cmd[1]=strdup("-c");
  cmd[2]=strdup(command);
  cmd[3]=NULL;

  message(1,"%s: MPI Serial %d: vfork -> sh -c \"%s\"\n",
	  timestamp(),mpi_rank,command);
  pid=vfork(); /* SPLIT PROCESS HERE */
  if(pid<0)
    message(1,"%s: MPI Serial %d: %s: ERROR: vfork failed: %s\n",
	    timestamp(),mpi_rank,command,strerror(errno));
  else if(pid==0)
    /* We are in the child process.  After vfork, only two functions
       can safely be called: execve and _exit.  The below line will
       run the command using /bin/sh, and exit with the exit status of
       execve if /bin/sh cannot be started. */
    _exit(execve("/bin/sh",cmd,environ));

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
  message(2,"%s: MPI Serial %d: %s: before main loop\n",
	  timestamp(),mpi_rank,command);
  do {
    checked=0;
    message(3,"%s: MPI Serial %d: %s: top of main loop\n",
	    timestamp(),mpi_rank,command);
    if(pid<0) {
      /* Call to vfork failed.  Abort will be handled later in this loop. */
      rc=255;
      message(-1,"%s: MPI Serial %d: %s: vfork failed.  Set rc=%d.\n",
	      timestamp(),mpi_rank,command,rc);
    } else if(running) {
      /* Child was running in last iteration of loop.  Is child still
	 running? */
      message(3,"%s: MPI Serial %d: %s: running; call waitpid.  Set rc=%d.\n",
	      timestamp(),mpi_rank,command,rc);
      pid2=waitpid(pid,&ret,WNOHANG);
      checked=1;
      rc=0; /* Abort if this is non-zero in any rank after the big IF
	       block below. */
      if(pid2==pid) {
	message(3,"%s: MPI Serial %d: %s: waitpid status=%d.\n",
		timestamp(),mpi_rank,command,ret);
	if(WIFEXITED(ret)) {
	  rc=(unsigned int)WEXITSTATUS(ret);
	  running=0;
	  message(1,"%s: MPI Serial %d: %s: exit %d\n",
		  timestamp(),mpi_rank,command,rc);
	} else if(WIFSIGNALED(ret)) {
	  rc=(unsigned int)(128+WTERMSIG(ret));
	  running=0;
	  message(1,"%s: MPI Serial %d: %s: signal %d\n",
		  timestamp(),mpi_rank,command,WTERMSIG(ret));
	}  else if(WIFSTOPPED(ret)) {
	  rc=(unsigned int)(128+WSTOPSIG(ret));
	  running=0;
	  message(1,"%s: MPI Serial %d: %s: stopped by signal %d\n",
		  timestamp(),mpi_rank,command,WSTOPSIG(ret));
	} else {
	  message(1,"%s: MPI Serial %d: %s: unrecognized return status %d "
		  "(not -1, not exited, not signalled, not stopped)\n",
		  timestamp(),mpi_rank,command,ret);
	  running=0;
	  rc=255;
	}
      } else if(pid<0) {
	message(-1,"%s: MPI Serial %d: %s: ABORT: waitpid failed: %s\n",
		timestamp(),mpi_rank,command,strerror(errno));
	rc=255;
      } else {
	message(3,"%s: MPI Serial %d: %s: no status update from waitpid\n",
		timestamp(),mpi_rank,command);
      }
    } else {
      message(3,"%s: MPI Serial %d: %s: not checking process; no longer running.\n",
	      timestamp(),mpi_rank,command);
    }

    send[0]=rc;
    send[1]=running;
    send[2]=checked;
    message(3,"%s: MPI Serial %d: %s: local: rc=%d running=%d checked=%d\n",
	    timestamp(),mpi_rank,command,send[0],send[1],send[2]);
    mpicall(3,MPI_Allreduce(send,recv,3,MPI_UNSIGNED,MPI_MAX,MPI_COMM_WORLD),
	    "using an mpi_allreduce to get the maximum return code in monitor loop");

    now=time(NULL);
    if(now-lastprint>300 && verbosity<=1) {
      checkpoint_message(1,rc,running,recv[0],recv[1],pid);
      lastprint=now;
    } else if(now-lastprint>30 && verbosity<=2) {
      checkpoint_message(2,rc,running,recv[0],recv[1],pid);
      lastprint=now;
    } else {
      message(3,"%s: MPI Serial %d: %s: allreduce: recv[0]=rc=%d recv[1]=running=%d recv[2]=checked=%d\n",
	      timestamp(),mpi_rank,command,recv[0],recv[1],recv[2]);
      checkpoint_message(3,rc,running,recv[0],recv[1],pid);
    }

    if(!recv[1]) {
      message(1,"%s: MPI Serial %d: no tasks running.  Should exit %d.\n",
	      timestamp(),mpi_rank,recv[0]);
      return recv[0];
    } else if(!recv[2]) {
      die("%s: MPI Serial %d: ASSERTION FAILURE: no tasks checked children, but children are running.\n",
	  timestamp(),mpi_rank);
    } else {
      message(3,"%s: MPI Serial %d: non-zero recv[1]=%d\n",
	      timestamp(),mpi_rank,recv[1]);
    }

    if(recv[0] && running && immediate_exit) {
      /* At least one rank failed and we are instructed to exit immediately. */
      message(-1,"%s: MPI Serial %d: %s: non-zero exit from another rank; kill child process.\n",
	      timestamp(),mpi_rank,command);
      if(running && pid>0) {
	kill(pid,SIGTERM);
	/* Wait 5 seconds for child to exit from SIGTERM */
	for(i=0;i<200 && running;i++) { /* 200*0.05=10 seconds */
	  pid2=waitpid(pid,&ret,WNOHANG);
	  if(pid2==pid && (WIFEXITED(ret) || WIFSIGNALED(ret)))
	    /* Child exited. */
	    return recv[0];
	  doublesleep(0.05);
	}
	if(running) {
	  /* Child ignored SIGTERM for 10 seconds.  Send SIGKILL and a
	     scary message. */
	  message(-1,"%s: MPI Serial %d: %s: DANGER: child did not exit from SIGTERM.  "
		  "Sending STOP, KILL, CONT.\n",timestamp(),mpi_rank,command);
	  kill(pid,SIGSTOP);
	  kill(pid,SIGKILL);
	  kill(pid,SIGCONT);
	}
      }
    }

    message(3,"%s: MPI Serial %d: %s: sleep 0.5 seconds recv={%d %d}\n",
	    timestamp(),mpi_rank,command,recv[0],recv[1]);
    doublesleep(0.5);
  } while(recv[1]); /* while any ranks are running - bottom of main loop*/
  message(1,"%s: MPI Serial %d: %s: exited loop.\n",
	  timestamp(),mpi_rank,command);
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
  
  if(rc>255||rc<0) rc=255;

  return rc;
}

unsigned int spmd_run(const int argc,const char **argv,const int rank) {
  /* Executes the specified command using the "run" function,
     returning the resulting error code */
  char *command=append_args(argc,argv);
  unsigned int rc;
  rc=run(command);
  free(command);
  return rc;
}

void hydra_workaround(void) {
  struct stat s;
  if(!fstat(0,&s) && S_ISDIR(s.st_mode)) {
    fprintf(stderr,"warning: stdin is a directory; reopening /dev/null.  This is a bug in the Hydra MPI.  Submit a ticket to your helpdesk.\n");
    int x=open("/dev/null",O_RDONLY);
    if(x>0)
      dup2(x,0);
    else
      fprintf(stderr,"error: could not open /dev/null: %s\n",strerror(errno));
  }
}

void set_comm_size_rank(int commsize,int rank) {
  /* Sets the SCR_COMM_COMM and SCR_COMM_RANK variables to the
     specified values */
  char buffer[300];
  sprintf(buffer,"%d",rank);
  setenv("SCR_COMM_RANK",buffer,1);
  sprintf(buffer,"%d",commsize);
  setenv("SCR_COMM_SIZE",buffer,1);
}

unsigned int mpmd_run(const int argc,const char **argv,const int rank) {
  /* Executes the specified command using the "run" function,
     returning the resulting error code */

  char *file,*command;
  size_t len,cmdlen;
  unsigned int icmdlen;
  const int lentag=3013, strtag=3014;
  int ilen;
  unsigned int rc;

  /* Determine name of command file */
  if(rank==0) {
    if( (file=getenv("SCR_CMDFILE")) && (len=strlen(file)) ) {
      /* We have a non-zero-length command file name in $SCR_CMDFILE */
      ilen=len;
    } else {
      /* We do not have a valid command file name so use the default */
      size_t slen=strlen(default_cmdfile)+1;
      if( !(file=malloc(slen)) )
        die("Unable to allocate %llu bytes: %s\n",
            (unsigned long long)slen,strerror(errno));
      strcpy(file,default_cmdfile);
      ilen=strlen(default_cmdfile);
    }
  }

  /* Read command file on root */
  if(rank==0) {
    FILE *f=fopen(file,"rt");
    char buf[max_command_len],*cret,*point;
    size_t buflen=0;
    int lastline,blankline;
    int irank,commsize;

    if(!f)
      die("%s: Unable to open command file $SCR_CMDFILE=\"%s\": %s\n",
          timestamp(),file,strerror(errno));

    mpicall(2,MPI_Comm_size(MPI_COMM_WORLD,&commsize),"getting size of MPI_COMM_WORLD");

    lastline=0;
    for(irank=0;irank<commsize;irank++) {
      blankline=1;

      /* Skip over blank lines and comment lines */
      while(blankline && !lastline) {
        cret=fgets(buf,max_command_len,f);
        if(cret!=buf) {
          if(feof(f)) {
            lastline=1;
            strcpy(buf,special_non_command);
          } else
            die("%s: I/O error reading command file \"%s\": %s\n",
		timestamp(),file,strerror(errno));
        }

        for(point=buf;*point;point++) {
          if(*point==' ' || *point=='\t')
            continue;
          else if(*point=='\n' || *point=='\r')
            break; /* end of line */
          else if(*point=='#')
            break;
          else {
            blankline=0;
            break;
          }
        }
      }

      /* Handle command files that are too short */
      if(lastline && irank!=commsize-1)
        die("%s: Not enough ranks in $SCR_CMDFILE=\"%s\" for %d tasks (only %d present)\n",
            timestamp(),file,commsize,irank+1);

      if(blankline)
        die("%s: Not enough ranks in $SCR_CMDFILE=\"%s\" for %d tasks (only %d present)\n",
            timestamp(),file,commsize,irank);

      for(point=buf;*point;point++) {
        if(*point=='\r' || *point=='\n') {
          *point='\0';
          break;
        }
      }

      cmdlen=point-buf+1;
      icmdlen=cmdlen;
      if(cmdlen!=icmdlen) {
        die("%s: Command has length %llu which is too long to represent the length with "
	    "a %d-byte integer.\n",timestamp(),cmdlen,sizeof(int));
      }

      if(irank==0) {
        command=(char*)malloc(cmdlen);
        memcpy(command,buf,cmdlen);
      } else {
        mpicall(2,MPI_Send(&icmdlen,1,MPI_INTEGER,irank,lentag,MPI_COMM_WORLD),
                "sending command length");
        mpicall(2,MPI_Send(buf,cmdlen,MPI_CHARACTER,irank,strtag,MPI_COMM_WORLD),
                "sending command");
      }
    }
  } else {
    mpicall(2,MPI_Recv(&icmdlen,1,MPI_INTEGER,0,lentag,MPI_COMM_WORLD,MPI_STATUS_IGNORE),
            "receiving command length from root");
    cmdlen=icmdlen;
    command=malloc(cmdlen);
    if(!command)
      die("Unable to allocate %llu bytes for rank %d command\n",
          (unsigned long long)cmdlen,rank);
    mpicall(2,MPI_Recv(command,cmdlen,MPI_CHARACTER,0,strtag,
                     MPI_COMM_WORLD,MPI_STATUS_IGNORE),
            "receiving command from root");
    command[cmdlen-1]='\0';
  }

  rc=run(command);
  free(command);
  return rc;
}

int main(int argc,char **argv) {
  int err,rank;
  unsigned int rc;
  int maxcode,mycode,spmd,mpmd,commsize;

  verbosity=scr_verbosity();

  mpicall(2,MPI_Init( &argc, &argv ),"calling MPI_Init");
  mpi_inited=1; /* indicate to "die" that MPI_Init was called */

  hydra_workaround(); /* make sure stdin is not a directory */
  lsf_signal_workaround(); /* when using LSF, restore all signals */

  /* Enable MPI error handling so that we can give intelligent error messages */
  mpicall(2,MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_RETURN),"initializing mpi error handler");

  /* Determine the rank and size so we can forward that to the child processes */
  mpicall(2,MPI_Comm_rank(MPI_COMM_WORLD,&rank),"determining mpi rank in MPI_COMM_WORLD");
  mpicall(2,MPI_Comm_size(MPI_COMM_WORLD,&commsize),"determining MPI_COMM_WORLD size");

  mpi_rank=rank; /* avoid duplicate calls by storing rank */

  /* Determine the mode: SPMD or MPMD */
  spmd=is_spmd(argc,rank);
  mpmd=!spmd;

  /* Set the $SCR_COMM_RANK and $SCR_COMM_SIZE environment variables */
  set_comm_size_rank(commsize,rank);

  message(1,"%s,MPI Serial %d: MPI_COMM_WORLD size=%d rank=%d\n",
          timestamp(),rank,commsize,rank);
  message(2,"%s: MPI Serial %d: high verbosity enabled (SCR_VERBOSITY=%d)\n",
          timestamp(),rank,verbosity);

  /* Run the command */
  if(spmd) {
    if(argc<2)
      die("%s: Incorrect format for SPMD mode.  Format: %s command to run\n"
	  "Will run the command on all MPI ranks and exit with highest error status.\n",
	  timestamp(),argv[0]);
    message(2,"%s: MPI Serial %d: SPMD mode.\n",timestamp(),mpi_rank);
    rc=spmd_run(argc,(const char**)argv,rank);
  } else {
    if(argc>1)
      die("%s: ERROR.  Do not specify arguments in MPMD mode.\n",timestamp());
    message(2,"%s: MPI Serial %d: MPMD mode.\n",timestamp(),mpi_rank);
    rc=mpmd_run(argc,(const char**)argv,rank);
  }

  /* Exit immediately if non-zero exit status and
     SCR_IMMEDIATE_EXIT=YES */
  if(rc && scr_immediate_exit()) {
    fprintf(stderr,"%s: %s: exit status was %d and SCR_IMMEDIATE_EXIT is enabled.  Calling MPI_Abort.\n",
	    timestamp(),argv[0],rc);
    mpicall(2,MPI_Abort(MPI_COMM_WORLD,255),"calling MPI_Abort");
    return 255;
  }
 
  /* Determine the maximum return code */
  mycode=rc;
  maxcode=255;
  message(2,"%s: MPI Serial %d: get max return code (local=%d)\n",
	  timestamp(),mpi_rank,mycode);
  mpicall(2,MPI_Allreduce(&mycode,&maxcode,1,MPI_INTEGER,MPI_MAX,MPI_COMM_WORLD),
	  "using an mpi_allreduce to get the maximum return code at end of program");
  message(1,"%s: MPI Serial %d: max return code is %d\n",
	  timestamp(),mpi_rank,maxcode);

#ifdef MPI_ABORT_FOR_NONZERO_EXIT
  if(mpi_rank==0 && maxcode!=0)  
    fprintf(stderr,"%s: %s: maximum exit status was %d.  Calling MPI_Abort instead "
	    "of exiting to work around a problem in your MPI implementation.\n",
	    timestamp(),argv[0],maxcode);
#endif /* MPI_ABORT_FOR_NONZERO_EXIT */
  
  /* This final MPI_Barrier is necessary on some platforms: */
  mpicall(2,MPI_Barrier(MPI_COMM_WORLD),"during final MPI_Barrier");

#ifdef MPI_ABORT_FOR_NONZERO_EXIT
  if(maxcode!=0) {
    /* Have to MPI_Abort in order to ensure a non-zero exit status due
       to SGI MPT bug. */
    mpicall(2,MPI_Abort(MPI_COMM_WORLD,maxcode),"calling MPI_Abort");
  } else
#endif /* MPI_ABORT_FOR_NONZERO_EXIT */

    mpicall(2,MPI_Finalize(),"finalizing MPI library");

  /* Should never reach this line */
  return maxcode;
}
